import os
import requests
import zipfile
import shutil
from enum import Enum

import regis.diagnostics
import regis.util
import regis.rex_json
import regis.task_raii_printing

root = regis.util.find_root()
settings = regis.rex_json.load_file(os.path.join(root, regis.util.settingsPathFromRoot))
temp_dir = os.path.join(root, settings["intermediate_folder"], 'tmp')

class Host(Enum):
    UNKNOWN = 0
    GITLAB = 1
    GITHUB = 2

def _get_host(path):
    if "gitlab" in path:
        return Host.GITLAB
    elif "github" in path:
        return Host.GITHUB
    
    regis.diagnostics.log_info("Unknown host!")
    return Host.UNKNOWN

def _build_gitlab_path(baseUrl, name, tag):
    url = os.path.join(baseUrl, "-")
    url = os.path.join(url, "archive")
    url = os.path.join(url, tag)
    url = os.path.join(url, name+"-"+tag+".zip")
    url = url.replace("\\", "/")
    return url
def _build_github_path(baseUrl, tag):
    url = os.path.join(baseUrl, "archive")
    url = os.path.join(url, "refs")
    url = os.path.join(url, "tags")
    url = os.path.join(url, tag+".zip")
    url = url.replace("\\", "/")
    return url
def _build_host_path(baseUrl, name, tag):
    host = _get_host(baseUrl)
    if host == Host.GITHUB:
        return _build_github_path(baseUrl, tag)
    elif host == Host.GITLAB:
        return _build_gitlab_path(baseUrl, name, tag)
    else:
        regis.diagnostics.log_err(f"Unknown url host: {host} in url: {baseUrl}")
        return ""

def _load_externals_required():
    json_blob = regis.rex_json.load_file(os.path.join(root, "_build", "config", "required_externals.json"))
    if json_blob == None:
        regis.diagnostics.log_err("Loaded json blob is None, stopping json parse")
        return []

    externals_required = []
    for object in json_blob:
        externals_required.append(json_blob[object])

    return externals_required

def _download_external(url):
    # create temporary directory to store cached files to
    if not os.path.exists(temp_dir):
        regis.diagnostics.log_info(f'creating: {temp_dir}')
        os.mkdir(temp_dir)

    # get basename of the URL (a.k.a. the filename + extention we would like to download)
    url_basename = os.path.basename(url)
    download_filepath = os.path.join(temp_dir, url_basename)

    # request a download of the given URL
    if not os.path.exists(download_filepath):
        response = requests.get(url)
        if response.status_code == requests.codes.ok:
            # write the downloaded file to disk
            open(download_filepath, "wb").write(response.content)
        else:
            # bad request was made
            regis.diagnostics.log_err(f"Bad request [{str(response.status_code)}] for given url: {url}")
            return []
        
    # extract the zip file on disk
    # we cache the files within the directory before 
    # and after extraction, this gives us the ability
    # to examine the added files within the directory
    regis.diagnostics.log_info("Extracting: " + url)
    
    # pre list directories
    # cached directories before we downloaded anything
    pre_list_dir = os.listdir(temp_dir)
    with zipfile.ZipFile(download_filepath,"r") as zip_ref:
        zip_ref.extractall(temp_dir)

    # post list directories
    # directories after we downloaded the repository
    post_list_dir = os.listdir(temp_dir)

    regis.diagnostics.log_info("Looking for added extracted directories ...")
    added_directory_names = []
    for post_dir in post_list_dir:
        count = pre_list_dir.count(post_dir)
        if count == 0:
            added_directory_names.append(post_dir)
    regis.diagnostics.log_info(f"Found ({str(len(added_directory_names))}): ".join(added_directory_names))

    # remove the created zip file
    os.remove(download_filepath)

    return added_directory_names

def _verify_external(externalPath, requiredTag):
    external_name = os.path.basename(externalPath)

    if os.path.exists(externalPath):
        regis.diagnostics.log_no_color(f"External found: {external_name}")
        regis.diagnostics.log_no_color(f"validating version ...")
        version = regis.util.load_version_file(externalPath)
        if version != requiredTag:
            regis.diagnostics.log_err(f"Invalid version data found, redownloading external: {external_name}")
            return False
        regis.diagnostics.log_info(f"correct version")
        return True

    else:
        return False

def _install_external(external):
    with regis.task_raii_printing.TaskRaiiPrint("Installing externals.."):
        external_url = external["url"]
        external_name = external["name"]
        external_tag = external["tag"]
        external_store = external["storage"]
        external_store = external_store.replace("~", root)

        externals_dir = os.path.join(external_store, external_name)

        # if the external is already present we need to check if we need to redownload anything
        valid_external = _verify_external(externals_dir, external_tag)
        if not valid_external:    
            # any data that was already available will be deleted 
            # the data will be out of date anyway when a download is triggered
            if os.path.exists(externals_dir):
                shutil.rmtree(externals_dir)

            url = _build_host_path(external_url, external_name, external_tag)
            added_directories = _download_external(url)     

            if len(added_directories) == 1:
                # move to output directory
                shutil.move(os.path.join(temp_dir, added_directories[0]), os.path.join(external_store, added_directories[0]))
                # change directory name
                cwd = os.getcwd()
                os.chdir(external_store)
                os.rename(added_directories[0], external_name)
                os.chdir(cwd)
            elif len(added_directories) > 1:
                # create output directory
                if not os.path.exists(externals_dir):
                    os.makedirs(externals_dir)
                # move to output directory
                for added_directory in added_directories:
                    shutil.move(os.path.join(temp_dir, added_directory), externals_dir)
            else:
                regis.diagnostics.log_err("No directories where extracted.")
                return

            regis.util.create_version_file(externals_dir, external_tag)   

def _remove_tmp_dir():
    if os.path.exists(temp_dir):
        shutil.rmtree(temp_dir)

def query():
    externals_required = _load_externals_required()
    if externals_required == None:
        regis.diagnostics.log_err("Required externals is None, exiting ...")
        return False
    
    res = True
    for external in externals_required:

        external_tag = external["tag"]
        external_name = external["name"]
        external_store = external["storage"]
        external_store = external_store.replace("~", root)
        externals_dir = os.path.join(external_store, external_name)

        res &= _verify_external(externals_dir, external_tag)

    return res

def install():
    externals_required = _load_externals_required()
    if externals_required == None:
        regis.diagnostics.log_err("Required externals is None, exiting ...")
        return

    for external in externals_required:
        _install_external(external)    

    _remove_tmp_dir()



import os
import time
import watchdog.observers as ob
from enum import Enum, auto
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler, RegexMatchingEventHandler, FileSystemEvent, FileSystemMovedEvent

ob.read_directory_changes.WATCHDOG_TRAVERSE_MOVED_DIR_DELAY = 0

class FileOperation(Enum):
    Open = 0
    Close = auto()
    Created = auto()
    Moved = auto()
    Modified = auto()
    Deleted = auto()

class Operation():
    def __init__(self, filepath : str, op : FileOperation, time : str):
        self.filepath = filepath
        self.op = op
        self.time = time

    def __str__(self):
        return f'[{self.time}][{self.op.name}] {self.filepath}'

class DirWatcher():
    def __init__(self, dir : str, bRecursive : bool):
        self.event_handler = FileSystemEventHandler()

        self.event_handler.on_created = self._on_created
        self.event_handler.on_deleted = self._on_deleted
        self.event_handler.on_modified = self._on_modified
        self.event_handler.on_moved = self._on_moved
        self.event_handler.on_opened = self._on_opened
        self.event_handler.on_closed = self._on_closed

        self.observer = Observer()
        self.observer.schedule(self.event_handler, dir, recursive=bRecursive)

        self.operations : list[Operation] = []

    def __enter__(self):
        self.observer.start()       
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.observer.stop()
        self.observer.join()

    def _on_created(self, event : FileSystemEvent):
        full_path = os.path.normpath(os.path.join(os.getcwd(), event.src_path))
        self.operations.append(Operation(full_path, FileOperation.Created, time.asctime()))
    
    def _on_deleted(self, event : FileSystemEvent):
        full_path = os.path.normpath(os.path.join(os.getcwd(), event.src_path))
        self.operations.append(Operation(full_path, FileOperation.Deleted, time.asctime()))
    
    def _on_modified(self, event : FileSystemEvent):
        full_path = os.path.normpath(os.path.join(os.getcwd(), event.src_path))
        self.operations.append(Operation(full_path, FileOperation.Modified, time.asctime()))
    
    def _on_moved(self, event : FileSystemMovedEvent):
        full_dest_path = os.path.normpath(os.path.join(os.getcwd(), event.dest_path))
        full_src_path = os.path.normpath(os.path.join(os.getcwd(), event.src_path))

        self.operations.append(Operation(f'{full_src_path} -> {full_dest_path}', FileOperation.Moved, time.asctime()))
        
    def _on_opened(self, event : FileSystemEvent):
        full_path = os.path.normpath(os.path.join(os.getcwd(), event.src_path))
        self.operations.append(Operation(full_path, FileOperation.Open, time.asctime()))

    def _on_closed(self, event : FileSystemEvent):
        full_path = os.path.normpath(os.path.join(os.getcwd(), event.src_path))
        self.operations.append(Operation(full_path, FileOperation.Close, time.asctime()))




import os
import sys
import regis.rex_json
import regis.util
import regis.task_raii_printing
import regis.diagnostics
import regis.vs
import requests
import zipfile 
import shutil
import threading
import argparse

from pathlib import Path

root = regis.util.find_root()
settings = regis.rex_json.load_file(os.path.join(root, regis.util.settingsPathFromRoot))
build_dir = os.path.join(root, settings["build_folder"])
temp_dir = os.path.join(root, settings["intermediate_folder"])
tools_install_dir = os.path.join(temp_dir, settings["tools_folder"])
tool_paths_filepath = os.path.join(tools_install_dir, "tool_paths.json")
zip_downloads_path = os.path.join(tools_install_dir, "zips")

tool_paths_dict = {}
if os.path.exists(tool_paths_filepath):
  tool_paths_dict = regis.rex_json.load_file(tool_paths_filepath)
required_tools = []
not_found_tools = []
found_tools = []

def __load_required_tools_dict():
  tools_required = []
  json_blob = regis.rex_json.load_file(os.path.join(root, "_build", "config", "required_tools.json"))
  for object in json_blob:
    tools_required.append(json_blob[object])

  return tools_required

def __print_tool_found(tool, path : str):
  regis.diagnostics.log_no_color(f"{tool['config_name']} found at {path}")

def __get_tool_extension(tool):
  extension = "" 
  if "extension" in tool:
    extension = tool["extension"]
  elif regis.util.is_windows():
    extension = ".exe"

  return extension

def __get_msvc_paths(msvcRelPath : str):
  res = []

  installed_vs_versions = regis.vs.installed_versions()
  tool_path = msvcRelPath
  for installed_vs in installed_vs_versions:
    msvc_path = os.path.join(installed_vs.installation_path, 'VC', 'Tools', tool_path)
    if os.path.exists(msvc_path):
      res.append(msvc_path)
    
  return res

def __get_possible_tool_install_dirs(requiredTool : dict):
  res = []

  # we purposely don't look in the PATH env variable
  # as we cannot confidently check version of tools
  # found in PATH.

  # some tools have specific locations they can be installed at
  if requiredTool['archive_name'] == 'MSVC':
    res.extend(__get_msvc_paths(requiredTool['path']))

  return res

def __look_for_tool_on_system(requiredTool):
    possible_paths = __get_possible_tool_install_dirs(requiredTool)

    # look for the tool
    exe_extension = __get_tool_extension(requiredTool)
    tool_path = regis.util.find_file_in_paths(f"{requiredTool['stem']}{exe_extension}", possible_paths)

    return tool_path

def __look_for_tools(required_tools):   
  for required_tool in required_tools:
    stem = required_tool["stem"]
    path = os.path.join(tools_install_dir, required_tool["archive_name"])
    version = regis.util.load_version_file(path)

    # If the version file is not found
    # we look for the tool somewhere else on the system
    if not version:
      tool_path = __look_for_tool_on_system(required_tool)
      # if the tool is found, we save this path instead
      # we don't download the tool from the web
      if tool_path != '':
        __print_tool_found(required_tool, tool_path)
        tool_config_name = required_tool["config_name"]
        tool_paths_dict[tool_config_name] = tool_path
        found_tools.append(required_tool)
        continue

      regis.diagnostics.log_err(f"{stem} not found")
      not_found_tools.append(required_tool)
      continue

    # check if our version is up to date
    if version != required_tool["version"]:
      regis.diagnostics.log_err(f"{stem} is out of date")
      not_found_tools.append(required_tool)
      continue

    config_name = required_tool["config_name"]

    # check if the tool path is already in the cached paths
    if config_name in tool_paths_dict:
      tool_path = tool_paths_dict[config_name]
      if (os.path.exists(tool_path)):
          regis.diagnostics.log_no_color(f"{stem} found at {tool_path}")
          continue
      else:
        regis.diagnostics.log_err(f"Error: tool path cached, but path doesn't exist: {tool_path}")
        not_found_tools.append(required_tool)
        continue

    # if not, add the path of the tool directory where it'd be downloaded to
    paths_to_use = []
    paths_to_use.append(os.path.join(tools_install_dir, required_tool["path"]))

    # look for the tool
    exe_extension = __get_tool_extension(required_tool)
    tool_path = regis.util.find_file_in_paths(f"{stem}{exe_extension}", paths_to_use)

    # tool is found, add it to the cached paths
    if tool_path != '':
      __print_tool_found(required_tool, tool_path)
      tool_config_name = required_tool["config_name"]
      tool_paths_dict[tool_config_name] = tool_path
      found_tools.append(required_tool)
    # tool is not found, add it to the list to be looked for later
    else:
      not_found_tools.append(required_tool)

  return not_found_tools

def _are_installed():
  with regis.task_raii_printing.TaskRaiiPrint("Checking if tools are installed"):

    global required_tools
    required_tools = __load_required_tools_dict()
    
    global tool_paths_dict
    if tool_paths_dict == None:
      tool_paths_dict = {}
      
    global not_found_tools
    not_found_tools = __look_for_tools(required_tools)

    if len(not_found_tools) == 0:
      regis.diagnostics.log_info("All tools found")
      regis.rex_json.save_file(tool_paths_filepath, tool_paths_dict)
      return True
    else:
      regis.diagnostics.log_warn(f"Tools that weren't found or were out of date: ")
      for tool in not_found_tools:
        regis.diagnostics.log_warn(f"\t-{tool['stem']}")

  return False

def __download_file(url):
  filename = os.path.basename(url)
  filePath = os.path.join(zip_downloads_path, filename)
  
  if not os.path.exists(filePath):
    response = requests.get(url)
    open(filePath, "wb").write(response.content)
  
def __make_zip_download_path():
  if not os.path.exists(zip_downloads_path):
    os.makedirs(zip_downloads_path)

def __download_tool(name, version, numZipFiles):
  with regis.task_raii_printing.TaskRaiiPrint(f"Downloading tool {name} {version}"):

    with regis.util.LoadingAnimation('Downloading'):

      threads = []
      for i in range(numZipFiles):
        threads.append(__launch_download_thread((f"https://github.com/RisingLiberty/RegisZip/raw/{version}/data/{name}.zip.{(i + 1):03d}")))

      for thread in threads:
        thread.join()

def __download_tools_archive():
  with regis.task_raii_printing.TaskRaiiPrint("Downloading tools"):

    # filter duplicate tools
    tools_to_download = []
    for not_found_tool in not_found_tools:
      archive_name = not_found_tool["archive_name"]
      should_add = True
      for tool_to_download in tools_to_download:
        if archive_name == tool_to_download["archive_name"]:
          should_add = False
          break
      
      if should_add:
        tools_to_download.append(not_found_tool)

    for not_found_tool in tools_to_download:
      arch_name = not_found_tool["archive_name"]
      num_zip_files = not_found_tool["num_zip_files"]
      version = not_found_tool["version"]
      __download_tool(arch_name, version, num_zip_files)

    return tools_to_download
    
def __enumerate_tools(zipsFolder):
  zips = os.listdir(zipsFolder)
  tools = []
  for zip in zips:
    stem = Path(zip).stem
    if stem not in tools:
      tools.append(stem)

  return tools

def __zip_files_for_tool(stem, folder):
  zips = os.listdir(folder)
  tool_zip_files = []
  for zip in zips:
    if Path(zip).stem == stem:
      tool_zip_files.append(os.path.join(folder, zip))

  return tool_zip_files

def __unzip_tools():
  with regis.task_raii_printing.TaskRaiiPrint("Unzipping files"):
    tools_to_unzip = __enumerate_tools(zip_downloads_path)

    with regis.util.LoadingAnimation('Extracting zips'):
      for tool in tools_to_unzip:
        tool_zip_files = __zip_files_for_tool(tool, zip_downloads_path)
        tool_master_zip = os.path.join(zip_downloads_path, f"{tool}")
        regis.diagnostics.log_info(f'extracting {tool} to {tool_master_zip}')
        with open(tool_master_zip, "ab") as f:
          for tool_zip in tool_zip_files:
            with open(tool_zip, "rb") as z:
                f.write(z.read())

        try:
          with zipfile.ZipFile(tool_master_zip, "r") as zip_obj:
              zip_obj.extractall(tools_install_dir)
        except zipfile.BadZipFile as ex:
          regis.diagnostics.log_err(f'Unable to extract {tool_master_zip}. {ex}')
          sys.exit(1)

      regis.diagnostics.log_info(f"tools unzipped to {tools_install_dir}")

def __create_version_files(foundTools : []):
  for tool in foundTools:
    path = os.path.join(tools_install_dir, tool["archive_name"])
    regis.util.create_version_file(path, tool["version"])

def __delete_tmp_folders():
  if os.path.isdir(zip_downloads_path):
    shutil.rmtree(zip_downloads_path)

def __launch_download_thread(url):
    thread = threading.Thread(target=__download_file, args=(url,))
    thread.start()
    return thread  

def _download():
  with regis.task_raii_printing.TaskRaiiPrint("Downloading tools"):
    __delete_tmp_folders() # clean up if last job failed
    __make_zip_download_path()
    __download_tools_archive()
    __unzip_tools()
    __delete_tmp_folders()

def _install():
  with regis.task_raii_printing.TaskRaiiPrint("installing tools"):

    global tool_paths_dict
    global not_found_tools
    for tool in not_found_tools:

      # look for tool in the folder where it'd be downloaded to
      exe_extension = __get_tool_extension(tool)
      path = regis.util.find_file_in_folder(f"{tool['stem']}{exe_extension}", os.path.join(tools_install_dir, tool["path"]))

      # if not found, something is wrong and we have to investigate manually
      if path == "":
        tool_name = tool["stem"]
        regis.diagnostics.log_err(f"failed to find {tool_name}")
      else:
        # if found, add it to the cached paths
        __print_tool_found(tool, path)
        tool_config_name = tool["config_name"]
        tool_paths_dict[tool_config_name] = path
    
    found_tools.extend(not_found_tools)

    __create_version_files(found_tools)

    # save cached paths to disk
    regis.rex_json.save_file(tool_paths_filepath, tool_paths_dict)

def query():
  """Query which required tools are still missing on the current machine."""
  return _are_installed()

def install():
  if not _are_installed():
    _download()
    _install()




import os
import sys
import threading
import regis.required_tools
import regis.rex_json
import regis.util
import regis.diagnostics
import regis.subproc
import regis.dir_watcher
import regis.generation

from pathlib import Path

from requests.structures import CaseInsensitiveDict

tool_paths_dict = regis.required_tools.tool_paths_dict
root = regis.util.find_root()  
settings_path = os.path.join(root, regis.util.settingsPathFromRoot)
settings = regis.rex_json.load_file(settings_path)
intermediate_path = os.path.join(root, settings['intermediate_folder'], settings['build_folder'])
build_projects_path = os.path.join(intermediate_path, settings['build_projects_filename'])

class NinjaProject:
  def __init__(self, filepath : str):
    self.json_blob : dict = regis.rex_json.load_file(filepath)
    self.filepath = filepath
    self.project_name = self.json_blob['name']

  def ninja_file(self, compiler : str, config : str):
    return self.json_blob['configs'][compiler.lower()][config.lower()]["ninja_file"]

  def dependencies(self, compiler : str, config : str):
    return self.json_blob['configs'][compiler.lower()][config.lower()]["dependencies"]

  def clean(self, compiler : str, config : str, buildDependencies : bool, verboseOutput = False):
    ninja_path = tool_paths_dict["ninja_path"]
    regis.diagnostics.log_info(f'Cleaning intermediates')
    
    r = self._valid_args_check(compiler, config)
    if r != 0:
      return r

    if buildDependencies:
      self._clean_dependencies(compiler, config, verboseOutput)

    cmd = f"{ninja_path} -f {self.ninja_file(compiler, config)} -t clean"

    if verboseOutput:
      cmd += ' -v'

    proc = regis.subproc.run(cmd)
    proc.wait()

    r = proc.returncode
    return r
  
  def build(self, compiler : str, config : str, buildDependencies : bool, verboseOutput = False):
    r = 0

    r = self._valid_args_check(compiler, config)
    if r != 0:
      return r

    # make sure to build the dependencies first
    if buildDependencies:
      r |= self._build_dependencies(compiler, config, verboseOutput)

    regis.diagnostics.log_info(f"Building: {self.project_name} - {config} - {compiler}")

    # then build the project we specified
    ninja_path = tool_paths_dict["ninja_path"]
    cmd = f"{ninja_path} -f {self.ninja_file(compiler, config)}"

    if verboseOutput:
      cmd += ' -v'

    regis.diagnostics.log_info(f'executing: {cmd}')
    proc = regis.subproc.run(cmd)
    proc.wait()
    r |= proc.returncode

    # show error if any build failed
    if r != 0:
      regis.diagnostics.log_err(f"Failed to build {self.project_name}")

    return r
  
  def rebuild(self, compiler : str, config : str, buildDependencies : bool):
    r = self.clean(compiler, config, buildDependencies)
    if r != 0:
      return r
    
    r = self.build(compiler, config, buildDependencies)

    return r
   
  def _valid_args_check(self, compiler : str, config : str):
    if compiler not in self.json_blob['configs']:
      regis.diagnostics.log_err(f"no compiler '{compiler}' found for project '{self.project_name}'")
      regis.diagnostics.log_err('Found compilers are')
      for compiler_name in self.json_blob['configs']:
        regis.diagnostics.log_err(f'- {compiler_name}')
      return 1
  
    if config not in self.json_blob['configs'][compiler]:
      regis.diagnostics.log_err(f"error in {self.filepath}")
      regis.diagnostics.log_err(f"no config '{config}' found in project '{self.project_name}' for compiler '{compiler}'")
      regis.diagnostics.log_err('Found configs are')
      for config_name in self.json_blob['configs'][compiler]:
        regis.diagnostics.log_err(f'- {config_name}')
      return 1
    
    return 0

  def _build_dependencies(self, compiler, config, verboseOutput):
    dependencies = self.json_blob['configs'][compiler][config]["dependencies"]

    r = 0
    for dependency in dependencies:      
      dependency_project_name = Path(dependency).stem
      regis.diagnostics.log_info(f'Building dependency: {self.project_name} -> {dependency_project_name}')

      dependency_project = NinjaProject(dependency)
      r |= dependency_project.build(compiler, config, buildDependencies=True, verboseOutput=verboseOutput)

    return r

  def _clean_dependencies(self, compiler, config, verboseOutput):
    dependencies = self.json_blob['configs'][compiler][config]["dependencies"]

    r = 0
    for dependency in dependencies:      
      dependency_project_name = Path(dependency).stem
      regis.diagnostics.log_info(f'Building dependency: {self.project_name} -> {dependency_project_name}')

      dependency_project = NinjaProject(dependency)
      r |= dependency_project.clean(compiler, config, buildDependencies=True, verboseOutput=verboseOutput)

    return r

  def compilers(self):
    return self.json_blob['configs']

def find_sln(directory):
  """Find the ninja solution in the specified directory"""
  dirs = os.listdir(directory)

  res = []

  for dir in dirs:
    full_path = os.path.join(directory, dir)
    if os.path.isfile(full_path) and Path(full_path).suffix == ".nsln":
      res.append(full_path)
    
  return res

def _find_ninja_project_file(slnFile : str, projectName : str):
  sln_jsob_blob = CaseInsensitiveDict(regis.rex_json.load_file(slnFile))
  
  if projectName not in sln_jsob_blob:
    regis.diagnostics.log_err(f"project '{projectName}' was not found in solution, have you generated it?")
    regis.diagnostics.log_err(f'found projects are')
    for project in sln_jsob_blob:
      regis.diagnostics.log_err(f'- {project}')
    return None
  
  project_file_path = sln_jsob_blob[projectName]    
  return NinjaProject(project_file_path)

def _launch_new_build(project : NinjaProject, config : str, compiler : str, shouldBuild : bool, shouldClean : bool, buildDependencies = False, verboseOutput = False):
  """Load the solution, look for the project in the solution and build it"""
  compiler_lower = compiler.lower()
  config_lower = config.lower()

  r = 0

  if shouldClean:
    r |= project.clean(compiler_lower, config_lower, buildDependencies, verboseOutput)

  if shouldBuild:
    r |= project.build(compiler_lower, config_lower, buildDependencies, verboseOutput)

  return r

def _look_for_sln_file_to_use(slnFile : str):
  """Look for the specified sln. Look for a sln in the root if no solution path is specified."""
  if slnFile == "":
    root = regis.util.find_root()
    sln_files = find_sln(root)

    if len(sln_files) > 1:
      regis.diagnostics.log_err(f'more than 1 nsln file was found in the cwd, please specify which one you want to use')
    
      for file in sln_files:
        regis.diagnostics.log_err(f'-{file}')
    
      return ""
    
    if len(sln_files) == 0:
      regis.diagnostics.log_err(f'no nlsn found in {root}')
      return ""

    slnFile = sln_files[0]
  elif not os.path.exists(slnFile):
    regis.diagnostics.log_err(f'solution path {slnFile} does not exist')
    return ""
  
  return slnFile

def _update_cleaned_projects(project : str, config : str, compiler : str, deletedProgram : str):
  """Update the build projects file and remove all the projects that have been cleaned"""
  build_projects = regis.rex_json.load_file(build_projects_path)

  project = project.lower()
  config = config.lower()
  compiler = compiler.lower()

  if project not in build_projects:
    return
  
  build_project = build_projects[project]
  if config not in build_project:
    return
  
  build_config = build_projects[project][config]
  if compiler not in build_config:
    return
  
  build_programs : list[str] = build_projects[project][config][compiler]

  if deletedProgram in build_programs:
    build_programs.remove(deletedProgram)

  regis.rex_json.save_file(build_projects_path, build_projects)

def _update_build_projects(project : str, config : str, compiler : str, createdProgram : str):
  """Update the build projects file and update the paths to new build projects"""
  build_projects = regis.rex_json.load_file(build_projects_path)

  project = project.lower()
  config = config.lower()
  compiler = compiler.lower()

  if project not in build_projects:
    build_projects[project] = {}
  
  build_project = build_projects[project]
  if config not in build_project:
    build_project[config] = {}
    
  build_config = build_project[config]
  if compiler not in build_config:
    build_config[compiler] = []

  build_projects[project][config][compiler].append(createdProgram)

  regis.rex_json.save_file(build_projects_path, build_projects)

def new_build(projectName : str, config : str, compiler : str, shouldBuild : bool = False, shouldClean : bool = False, slnFile : str = "", buildDependencies : bool = False, verboseOutput : bool = False):
  """This is the interface to the build pipeline.\n
  It'll launch a new build for the project using the config and compiler specified.\n
  It's possible to to negate building and only clean or to do a clean step before the build starts.\n
  It's also possible to only build the project and not its dependencies"""  
  slnFile = _look_for_sln_file_to_use(slnFile)

  if slnFile == "":
    regis.diagnostics.log_err("unable to find solution, aborting..")
    return 1
  
  project = _find_ninja_project_file(slnFile, projectName)
  if not project:
    regis.diagnostics.log_err(f'Failed to find {projectName} in solution')
    sys.exit(1)

  with regis.dir_watcher.DirWatcher(intermediate_path, bRecursive=True) as dir_watcher:
    res = _launch_new_build(project, config, compiler, shouldBuild, shouldClean, buildDependencies, verboseOutput)

  if not os.path.exists(build_projects_path):
    regis.rex_json.save_file(build_projects_path, {})

  # it's possible nothing gets done because everything is up to date
  # in that case, we don't need to update anything
  for op in dir_watcher.operations:
    if regis.util.is_executable(op.filepath):
      if (op.op == regis.dir_watcher.FileOperation.Deleted):
        _update_cleaned_projects(projectName, config, compiler, op.filepath)
      elif (op.op == regis.dir_watcher.FileOperation.Created):
        _update_build_projects(projectName, config, compiler, op.filepath)

  return res
  
def build_all_configs(projectName : str, shouldBuild : bool = False, shouldClean : bool = False, slnFile : str = "", buildDependencies : bool = False, singleThreaded : bool = True, verboseOutput : bool = False):
  slnFile = _look_for_sln_file_to_use(slnFile)

  if slnFile == "":
    regis.diagnostics.log_err("unable to find solution, aborting..")
    return 1
  
  project = _find_ninja_project_file(slnFile, projectName)
  compilers = project.compilers()

  # This is the list that'll store the results of each build
  result_arr = []
  def _build_on_thread(prj, cfg, comp, result):
    result.append(_launch_new_build(prj, cfg, comp, shouldBuild, shouldClean, buildDependencies, verboseOutput))

  threads : list[threading.Thread] = []
  
  # loop over the configs and compilers and create a build for each combination
  for compiler in compilers:
    for config in compilers[compiler]:
      thread = threading.Thread(target=_build_on_thread, args=(project, config, compiler, result_arr))
      thread.start()

      # A dirty hack for singlethreaded mode
      # we always spawn a thread but in single threaded mode, we join it immediately
      if singleThreaded:
        thread.join()
      else:
        threads.append(thread)

  # in multi threaded mode, we join threads after all of them have spawned
  for thread in threads:
    thread.join()

  return result_arr.count(0) != len(result_arr)


# ============================================ 
#
# Author: Nick De Breuck
# Twitter: @nick_debreuck
# 
# File: test.py
# Copyright (c) Nick De Breuck 2023
#
# ============================================

import os
import threading
import time
import threading
import subprocess
import re
import shutil
import regis.required_tools
import regis.util
import regis.task_raii_printing
import regis.rex_json
import regis.code_coverage
import regis.diagnostics
import regis.generation
import regis.build
import regis.dir_watcher

from pathlib import Path
from datetime import datetime
from requests.structures import CaseInsensitiveDict
from enum import Enum, auto

root_path = regis.util.find_root()
tool_paths_dict = regis.required_tools.tool_paths_dict
settings = regis.rex_json.load_file(os.path.join(root_path, regis.util.settingsPathFromRoot))
_pass_results = {}

iwyu_intermediate_dir = "iwyu"
clang_tidy_intermediate_dir = "clang_tidy"
unit_tests_intermediate_dir = "unit_tests"
coverage_intermediate_dir = "coverage"
asan_intermediate_dir = "asan"
ubsan_intermediate_dir = "ubsan"
fuzzy_intermediate_dir = "fuzzy"
auto_test_intermediate_dir = "auto_test"

def get_pass_results():
  return _pass_results

def _is_in_line(line : str, keywords : list[str]):
  regex = "((error).(cpp))|((error).(h))"

  for keyword in keywords:
    if keyword.lower() in line.lower():
      return not re.search(regex, line.lower()) # make sure that lines like 'error.cpp' don't return positive

  return False

def _symbolic_print(line, filterLines : bool = False):
  error_keywords = ["failed", "error"]
  warn_keywords = ["warning"]

  if _is_in_line(line, error_keywords):
    regis.diagnostics.log_err(line)
  elif _is_in_line(line, warn_keywords):
    regis.diagnostics.log_warn(line)
  elif not filterLines:
    regis.diagnostics.log_no_color(line)

def _default_output_callback(pid, output, isStdErr, filterLines):
  logs_dir = os.path.join(settings["intermediate_folder"], settings["logs_folder"])
  filename = f"output_{pid}.log"
  if isStdErr:
    filename = f"errors_{pid}.log"

  filepath = os.path.join(logs_dir, filename)
  if os.path.exists(filepath):
    os.remove(filepath)
  elif not os.path.exists(logs_dir):
    os.makedirs(logs_dir)

  with open(filepath, "a+") as f:

    for line in iter(output.readline, b''):
      new_line : str = line.decode('UTF-8')
      if new_line.endswith('\n'):
        new_line = new_line.removesuffix('\n')

      _symbolic_print(new_line, filterLines)      
      f.write(f"{new_line}\n")

    regis.diagnostics.log_info(f"full output saved to {filepath}")

def _get_coverage_rawdata_filename(program : str):
  return f"{Path(program).stem}.profraw"

def _create_coverage_report(program, indexedFile):
  with regis.task_raii_printing.TaskRaiiPrint("creating coverage reports"):

    if Path(program).stem != Path(indexedFile).stem:
      regis.diagnostics.log_err(f"program stem doesn't match coverage file stem: {Path(program).stem} != {Path(indexedFile).stem}")
      return 1

    regis.code_coverage.create_line_oriented_report(program, indexedFile)
    regis.code_coverage.create_file_level_summary(program, indexedFile)
    regis.code_coverage.create_lcov_report(program, indexedFile)

  return 0

def _parse_coverage_report(indexedFile):
  with regis.task_raii_printing.TaskRaiiPrint("parsing coverage reports"):
      report_filename = regis.code_coverage.get_file_level_summary_filename(indexedFile)
      return regis.code_coverage.parse_file_summary(report_filename)

class RunnableType(Enum):
  Default = 0,
  Coverage = auto(),
  Sanitizer = auto(),

class Runnable():
  def __init__(self, runnableDict, args = [], enableAsan : bool = False, enableUbsan : bool = False):
    self.program = runnableDict['Program']
    self.type = RunnableType[runnableDict['RunnableType']]
    self.args = args
    self.proc = None
    self.terminated = False
    self.finished = False
    self.enable_asan = enableAsan
    self.enable_ubsan = enableUbsan
  
  def run(self):
    regis.diagnostics.log_info(f"running: {Path(self.program).name}")
    regis.diagnostics.log_info(f"with args: {self.args}")

    # error code by default in case we don't trigger any runnable
    rc = 1

    if self.type == RunnableType.Default:
      rc = self._default_run()
    
    if self.type == RunnableType.Coverage:
      rc = self._run_coverage()
    
    if self.type == RunnableType.Sanitizer:
      rc = self._run_sanitizer()
            
    self.finished = True
    return rc

  def terminate(self):
    if self.proc:
      self.proc.terminate()

    self.terminated = True
    self.finished = True

  def _default_run(self):
    self.proc = regis.util.run_subprocess(self.program, self.args)
    return regis.util.wait_for_process(self.proc)

  def _run_coverage(self):
    # First run the program
    coverage_rawdata_filename = _get_coverage_rawdata_filename(self.program)
    raw_data_file = os.path.join(Path(self.program).parent, coverage_rawdata_filename)
    os.environ["LLVM_PROFILE_FILE"] = raw_data_file # this is what llvm uses to set the raw data filename for the coverage data
    self.proc = regis.util.run_subprocess(self.program, self.args)
    rc = regis.util.wait_for_process(self.proc)

    # Then index the raw data file
    indexed_file = regis.code_coverage.create_index_rawdata(raw_data_file)

    # next create the coverage report
    rc |= _create_coverage_report(self.program, indexed_file)

    # finally, parse the coverage report
    rc |= _parse_coverage_report(indexed_file)

    return rc
  
  def _run_sanitizer(self):
    rc = 0

    # ASAN_OPTIONS common flags: https://github.com/google/sanitizers/wiki/SanitizerCommonFlags
    # ASAN_OPTIONS flags: https://github.com/google/sanitizers/wiki/AddressSanitizerFlags
    # UBSAN_OPTIONS common flags: https://github.com/google/sanitizers/wiki/SanitizerCommonFlags
    log_folder = os.path.join(root_path, settings["intermediate_folder"], settings["logs_folder"])
    
    asan_log_path = ''
    if self.enable_asan:
      asan_log_path = os.path.join(log_folder, 'asan.log').replace('\\', '/')
      asan_options = f"print_stacktrace=1:log_path=\"{asan_log_path}\""
      os.environ["ASAN_OPTIONS"] = asan_options # print callstacks and save to log file

    ubsan_log_path = ''
    if self.enable_ubsan:
      ubsan_log_path = os.path.join(log_folder, 'ubsan.log').replace('\\', '/')
      ubsan_options = f"print_stacktrace=1:log_path=\"{ubsan_log_path}\""
      os.environ["UBSAN_OPTIONS"] = ubsan_options # print callstacks and save to log file
    
    self.proc = regis.util.run_subprocess(self.program, self.args)
    new_rc = regis.util.wait_for_process(self.proc)
    if new_rc != 0 or (asan_log_path and os.path.exists(asan_log_path)) or (ubsan_log_path and os.path.exists(ubsan_log_path)):
      regis.diagnostics.log_err(f"sanitization failed for {self.program}") # use full path to avoid ambiguity
      regis.diagnostics.log_err(f"for more info regarding asan, please check: {asan_log_path}")
      regis.diagnostics.log_err(f"for more info regarding ubsan, please check: {ubsan_log_path}")
      new_rc = 1
    rc |= new_rc

    return rc

# ---------------------------------------------
# Code Analysis jobs
# ---------------------------------------------

class IncludeWhatYouUseJob():
  """A job that runs include-what-you-use over a project"""
  def __init__(self, shouldClean : bool, fixIncludes : bool):
    self.should_clean = shouldClean
    self.fix_includes = fixIncludes
    return
  
  def execute(self, singleThreaded : bool):
    regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
    rc = self._generate()
    _pass_results["include-what-you-use generate"] = rc
    if rc != 0:
      regis.diagnostics.log_err(f"include-what-you-use generation failed")
      return rc

    regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
    rc = self._run(singleThreaded)
    _pass_results["include-what-you-use run"] = rc

    if rc != 0:
      regis.diagnostics.log_err(f"include-what-you-use pass failed")
      return rc
    
    return rc
  
  def _generate(self):
    """Generate the projects to run include-what-you-use on"""
    config = regis.generation.create_config(f'-intermediate-dir={iwyu_intermediate_dir} -enable-clang-tools -disable-clang-tidy-for-thirdparty')
    return _generate_test_files(self.should_clean, iwyu_intermediate_dir, config)
  
  def _run(self, singleThreaded : bool):
    """Run include what you use on the codebase"""
    def _run(iwyuPath, compdb, outputPath, impPath, lock):
      """Run the actual include-what-you-use command and save the output into a log file"""
      # create the command line to launch include-what-you-use
      cmd = ""
      cmd += f"py {iwyuPath} -v -p={compdb}"
      cmd += f" -- -Xiwyu --quoted_includes_first"
      
      if impPath != "" and os.path.exists(impPath):
        cmd += f" -Xiwyu --mapping_file={impPath}"

      # run include-what-you-use and save the output into a file
      output = subprocess.getoutput(cmd)
      with open(outputPath, "w") as f:
        f.write(output)
      output_lines = output.split('\n')

      # print the output using our color coding
      # to detect if it's an error, warning or regular log
      with lock:
        for line in output_lines:
          _symbolic_print(line)

        # log to the user that output has been saved
        regis.diagnostics.log_info(f"include what you use info saved to {outputPath}")
  
    with regis.task_raii_printing.TaskRaiiPrint("running include-what-you-use"):
      # find all the compiler dbs.
      # these act as input for include-what-you-use
      intermediate_folder = os.path.join(root_path, settings["intermediate_folder"], settings["build_folder"], iwyu_intermediate_dir)
      result = regis.util.find_all_files_in_folder(intermediate_folder, "compile_commands.json")
        
      threads : list[threading.Thread] = []
      output_files_per_project : dict[str, list] = {}
      lock = threading.Lock()

      iwyu_path = tool_paths_dict["include_what_you_use_path"]
      iwyu_tool_path = os.path.join(Path(iwyu_path).parent, "iwyu_tool.py")

      # create the include-what-you-use jobs
      for compiler_db in result:
        compiler_db_folder = Path(compiler_db).parent
        impPath = os.path.join(compiler_db_folder, "iwyu.imp")
        output_path = os.path.join(compiler_db_folder, "iwyu_output.log")
        project_name = _get_project_name_of_compdb(compiler_db_folder)

        # if we haven't compiled this project yet, prep the dict for it
        if project_name not in output_files_per_project:
          output_files_per_project[project_name] = []

        output_files_per_project[project_name].append(output_path)

        thread = threading.Thread(target=_run, args=(iwyu_tool_path, compiler_db, output_path, impPath, lock))
        thread.start()

        # very simple way of splitting single threaded with multi threaded mode
        # if we're single thread, we create a thread and immediately join it
        # if we're using multi threaded mode, we join them after all of them have been created
        if singleThreaded:
          thread.join() 
        else:
          threads.append(thread)

      # join all the threads after they've all been created
      # this is basically a no op if we're running in single threaded mode
      for thread in threads:
        thread.join()

      threads.clear()

      # because different configs could require different symbols or includes
      # we need to process all configs first, then process each output file for each config
      # for a given project and only if an include is not needed in all configs
      # take action and remove it or replace it with a forward declare
      # this can't be multithreaded
      if self.fix_includes:
        regis.diagnostics.log_info(f'Applying fixes..')

      fix_includes_path = os.path.join(Path(iwyu_path).parent, "fix_includes.py")

      # this is the actual run in trying to fix the includes
      # however it can be faked when self.fix_includes is false
      # if so, it'll do a dry run without changing anything
      # it'll still return a proper return code
      # indicating if anything needs to be changed
      rc = 0
      for key in output_files_per_project.keys():
        output_files = output_files_per_project[key]
        lines = []
        regis.diagnostics.log_info(f'processing: {key}')

        # include-what-you-use uses the output path of iwyu to determine what needs to be fixed
        # we merge all the outputs of all runs of iwyu on a project in different configs
        # and pass that temporary file over to include what you use
        for file in output_files:
          f = open(file, "r")
          lines.extend(f.readlines())

        filename = f'{key}_tmp.iwyu'
        filepath = os.path.join(intermediate_folder, filename)
        f = open(filepath, "w")
        f.writelines(lines)
        f.close()

        # create the fix includes command line
        cmd = f"py {fix_includes_path} --noreorder --process_merged=\"{filepath}\" --nocomments --nosafe_headers"

        if self.fix_includes == False:
          cmd += f" --dry_run"

        # run the fix includes command line
        rc |= os.system(f"{cmd}")  

      return rc

class ClangTidyJob():
  """A job that runs clang-tidy over a project"""
  def __init__(self, shouldClean : bool, autoFix : bool, filterLines : bool, filesRegex : str):
    self.should_clean = shouldClean
    self.auto_fix = autoFix
    self.filter_lines = filterLines
    self.files_regex = filesRegex
    return
  
  def execute(self, singleThreaded : bool):
    regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
    rc = self._generate()
    _pass_results["clang-tidy generation"] = rc
    if rc != 0:
      regis.diagnostics.log_err(f"clang-tidy pass failed")
      return rc

    regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
    rc = self._run(self.filter_lines, singleThreaded)
    _pass_results["clang-tidy run"] = rc
    if rc != 0:
      regis.diagnostics.log_err(f"clang-tidy pass failed")
      return rc

    return rc

  def _generate(self):
    config = regis.generation.create_config(f'-intermediate-dir={clang_tidy_intermediate_dir} -enable-clang-tools -disable-clang-tidy-for-thirdparty')
    return _generate_test_files(self.should_clean, clang_tidy_intermediate_dir, config)
    
  def _run(self, filterLines : bool, singleThreaded : bool):
    """Run clang-tidy on the codebase"""
    rc = [0]
    def _run(cmd : str, rc : int):
      """Run the actual clang-tidy command"""
      regis.diagnostics.log_info(f"executing: {cmd}")
      proc = regis.util.run_subprocess_with_callback(cmd, _default_output_callback, filterLines)
      new_rc = regis.util.wait_for_process(proc)
      if new_rc != 0:
        regis.diagnostics.log_err(f"clang-tidy failed for {compiler_db}")
        regis.diagnostics.log_err(f"config file: {config_file_path}")
      rc[0] |= new_rc

    with regis.task_raii_printing.TaskRaiiPrint("running clang-tidy"):

      # get the compiler dbs that are just generated
      result = _find_files(_create_full_intermediate_dir(clang_tidy_intermediate_dir), lambda file: 'compile_commands.json' in file)

      # create the clang-tidy jobs, we limit ourselves to 5 threads at the moment as running clang-tidy is quite performance heavy
      threads : list[threading.Thread] = []
      threads_to_use = 5
      script_path = os.path.dirname(__file__)
      clang_tidy_path = tool_paths_dict["clang_tidy_path"]
      clang_apply_replacements_path = tool_paths_dict["clang_apply_replacements_path"]

      for compiler_db in result:
        compiler_db_folder = Path(compiler_db).parent
        config_file_path = f"{compiler_db_folder}/.clang-tidy_second_pass"

        project_name = _get_project_name_of_compdb(compiler_db_folder)
        header_filters = regis.util.retrieve_header_filters(compiler_db_folder, project_name)
        header_filters_regex = regis.util.create_header_filter_regex(header_filters)
        
        # build up the clang-tidy command
        cmd = f"py \"{script_path}/run_clang_tidy.py\""
        cmd += f" -clang-tidy-binary=\"{clang_tidy_path}\""  # location of clang-tidy executable
        cmd += f" -clang-apply-replacements-binary=\"{clang_apply_replacements_path}\"" # location of clang-apply-replacements executable
        cmd += f" -config-file=\"{config_file_path}\"" # location of clang-tidy config file
        cmd += f" -p=\"{compiler_db_folder}\"" # location of compiler db folder (not the location to the file, but to its parent folder)
        cmd += f" -header-filter={header_filters_regex}" # only care about headers of the current project
        cmd += f" -quiet" # we don't want extensive logging
        cmd += f" -j={threads_to_use}" # only use a certain amount of threads, to reduce the performance overhead

        # auto fix found issues. This doesn't work for every enabled check.
        if self.auto_fix:
          cmd += f" -fix"

        # add the regex of the files we care about
        cmd += f" {self.files_regex}"

        # perform an incremental run, avoid rescanning previous scanned files that didn't change (ignores cpp files if their headers changed)
        if not self.should_clean:
          cmd += f" -incremental"

        # dirty hack to enable single thread mode vs multi threaded mode
        # in single threaded mode, we join the threads immediately
        thread = threading.Thread(target=_run, args=(cmd,rc,))
        thread.start()

        if singleThreaded:
          thread.join()
        else:
          threads.append(thread)

      for thread in threads:
        thread.join()

      return rc[0]

# ---------------------------------------------
# Unit Tests
# ---------------------------------------------
#
# Supports asan, ubsan and code coverage
#
class UnitTestJob():
  """A job that runs unit test projects"""
  def __init__(self, projects : list[str], shouldClean : bool, enableAsan : bool, enableUbsan : bool, enableCodeCoverage : bool):
    self.projects = projects
    self.enable_asan = enableAsan
    self.enable_ubsan = enableUbsan
    self.enable_code_coverage = enableCodeCoverage
    self.should_clean = shouldClean
  
  def execute(self, singleThreaded : bool):
    regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")

    # Generate the unit tests
    rc = self._generate()
    _pass_results["unit tests generation"] = rc
    if rc != 0:
      regis.diagnostics.log_err(f"failed to generate tests")
      return rc

    # pull out the generated projects, so we know what we can build and run
    test_projects_path = os.path.join(root_path, settings['intermediate_folder'], settings['build_folder'], 'test_projects.json')
    if not os.path.exists(test_projects_path):
      regis.diagnostics.log_err(f'"{test_projects_path}" does not exist.')
      return rc | 1

    # if no projects are specified, we run on all of them
    test_projects = regis.rex_json.load_file(test_projects_path)
    unit_test_projects = CaseInsensitiveDict(test_projects["TypeSettings"].get("UnitTest"))

    self.projects = self.projects or list(unit_test_projects.keys())

    # If we still have no projects, generation silently failed.
    # either way, we need to exit here
    if not self.projects:
      regis.diagnostics.log_warn(f'No unit test projects found. have you generated them?')
      _pass_results["unit tests - nothing to do"] = rc
      return rc

    # Now build the projects we're interested in
    regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
    rc |= self._build(self.projects, singleThreaded)

    # if any builds fail, we can't run any tests
    # so we exit here
    _pass_results["unit tests building"] = rc
    if rc != 0:
      regis.diagnostics.log_err(f"failed to build tests")
      return rc

    # Now that we've build everything, let's run everything
    regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
    for project in self.projects:
      if project not in unit_test_projects:
        regis.diagnostics.log_err(f'project "{project}" not found in {test_projects_path}. Please check its generation settings')
        continue

      # get the project test settings out of our testing files
      project_settings = unit_test_projects[project]
      runnables = project_settings['TargetRunnables']
      working_dir = project_settings['WorkingDir']
      
      # run all the tests
      rc = self._run(runnables, working_dir)
      _pass_results[f"unit tests result - {project}"] = rc

    # Report any issues
    if rc != 0:
      regis.diagnostics.log_err(f"unit tests failed")
      return rc
    
    return rc    

  def _generate(self):
    with regis.task_raii_printing.TaskRaiiPrint("generating unit test projects"):
      config_args = []
      config_args.append(f'-intermediate-dir={unit_tests_intermediate_dir}')
      config_args.append('-disable-clang-tidy-for-thirdparty')
      config_args.append('-enable-unit-tests')
      config_args.append('-disable-default-generation')

      if self.enable_code_coverage:
        config_args.append('-enable-code-coverage')
        config_args.append('-disable-default-configs')

      if self.enable_asan:
        config_args.append('-enable-address-sanitizer')
        config_args.append('-disable-default-configs')

      if self.enable_ubsan:
        config_args.append('-enable-ub-sanitizer')
        config_args.append('-disable-default-configs')

      config = regis.generation.create_config(' '.join(config_args))
      return _generate_test_files(self.should_clean, unit_tests_intermediate_dir, config)
  
  def _build(self, projects : list[str], singleThreaded : bool):
    with regis.task_raii_printing.TaskRaiiPrint("building unit tests"):
      return _build_files(projects, singleThreaded)
  
  def _run(self, runnables : list, workingDir : str):
    with regis.task_raii_printing.TaskRaiiPrint("running unit tests"):
      with regis.util.temp_cwd(workingDir):
        rc = 0
    
        # loop over each unit test program path and run it
        for runnable_dict in runnables:
          runnable = Runnable(runnable_dict, [], self.enable_asan, self.enable_ubsan)
          new_rc = runnable.run()

          if new_rc != 0:
            regis.diagnostics.log_err(f"unit test failed for {runnable.program}") # use full path to avoid ambiguity
          rc |= new_rc

        return rc

# ---------------------------------------------
# Auto Tests
# ---------------------------------------------
#
# Supports asan, ubsan and code coverage
#

class AutoTestJob():
  """A job that runs auto tests"""
  def __init__(self, projects : list[str], timeoutInSeconds : int, shouldClean : bool, enableAsan : bool, enableUbsan : bool, enableCodeCoverage : bool):
    self.projects = projects
    self.enable_asan = enableAsan
    self.enable_ubsan = enableUbsan
    self.enable_code_coverage = enableCodeCoverage
    self.should_clean = shouldClean
    self.timeout_in_seconds = timeoutInSeconds
  
  def execute(self, singleThreaded : bool):
    regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")

    # Generate the unit tests
    rc = self._generate()
    _pass_results["auto testing generation"] = rc
    
    if rc != 0:
      regis.diagnostics.log_err(f"failed to generate auto test code")
      return rc
    
    # pull out the generated projects, so we know what we can build and run
    test_projects_path = os.path.join(root_path, settings['intermediate_folder'], settings['build_folder'], 'test_projects.json')
    if not os.path.exists(test_projects_path):
      regis.diagnostics.log_err(f'"{test_projects_path}" does not exist.')
      return rc | 1

    # if no projects are specified, we run on all of them
    test_projects = regis.rex_json.load_file(test_projects_path)
    auto_test_projects = CaseInsensitiveDict(test_projects["TypeSettings"].get("AutoTest"))

    self.projects = self.projects or list(auto_test_projects.keys())

    # If we still have no projects, generation silently failed.
    # either way, we need to exit here
    if not self.projects:
      regis.diagnostics.log_warn(f'No auto test projects found. have you generated them?')
      _pass_results["auto testing - nothing to do"] = rc
      return rc

    # Now build the projects we're interested in
    regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
    rc |= self._build(self.projects, singleThreaded)

    # if any builds fail, we can't run any tests
    # so we exit here
    _pass_results["auto testing building"] = rc
    if rc != 0:
      regis.diagnostics.log_err(f"failed to build auto test code")
      return rc
    
    # Now that we've build everything, let's run everything
    for project in self.projects:
      if project not in auto_test_projects:
        regis.diagnostics.log_err(f'project "{project}" not found in {test_projects_path}. Please check its generation settings')
        continue

      # get the project test settings out of our testing files
      project_settings = auto_test_projects[project]
      runnables = project_settings['TargetRunnables']
      working_dir = project_settings['WorkingDir']
      test_file = _find_tests_file(project_settings)

      # run all the tests
      new_rc = self._run(runnables, working_dir, test_file, self.timeout_in_seconds)
      _pass_results[f'auto tests result - {project}'] = new_rc

      rc |= new_rc

    # report any issues
    if rc != 0:
      regis.diagnostics.log_err(f"auto tests failed")
      return rc

    return rc
  
  def _generate(self):
    with regis.task_raii_printing.TaskRaiiPrint("generating auto tests"):
      config_args = []
      config_args.append(f'-intermediate-dir={auto_test_intermediate_dir}')
      config_args.append('-enable-auto-tests')
      config_args.append('-disable-default-generation')

      if self.enable_code_coverage:
        config_args.append('-enable-code-coverage')
        config_args.append('-disable-default-configs')

      if self.enable_asan:
        config_args.append('-enable-address-sanitizer')
        config_args.append('-disable-default-configs')

      if self.enable_ubsan:
        config_args.append('-enable-ub-sanitizer')
        config_args.append('-disable-default-configs')

      config = regis.generation.create_config(' '.join(config_args))
      return _generate_test_files(self.should_clean, auto_test_intermediate_dir, config)
  
  def _build(self, projects : list[str], singleThreaded : bool):
    return _build_files(projects, singleThreaded)
  
  def _run(self, runnables : list[str], workingDir : str, testFilePath : str, timeoutInSeconds : int):
    json_blob = regis.rex_json.load_file(testFilePath)

    with regis.task_raii_printing.TaskRaiiPrint("running auto tests"):
      with regis.util.temp_cwd(workingDir):
        rc = 0

        for test in json_blob:
          command_line : str = json_blob[test]["command_line"]
          now = time.time()
          max_seconds = timeoutInSeconds
          def monitor_runnable(runnable : Runnable):
            while not runnable.finished:
              duration = time.time() - now
              if duration > max_seconds:
                runnable.terminate()

              time.sleep(1)

          # loop over each unit test program path and run it
          for runnable_dict in runnables:
            runnable = Runnable(runnable_dict, command_line, self.enable_asan, self.enable_ubsan)
            thread = threading.Thread(target=monitor_runnable, args=(runnable,))
            thread.start()

            new_rc = runnable.run()
            thread.join()

            if new_rc != 0:
              if runnable.terminated:
                regis.diagnostics.log_err(f"auto test timeout triggered for {runnable.program} after {max_seconds} seconds") # use full path to avoid ambiguity
              else:
                rc |= new_rc
                regis.diagnostics.log_err(f"auto test failed for {runnable.program} with returncode {new_rc}") # use full path to avoid ambiguity
          
        return rc
  
# ---------------------------------------------
# Fuzzy Tests
# ---------------------------------------------
#
# Supports asan, ubsan and code coverage
#

class FuzzyTestJob():
  """A job that runs fuzzy tests"""
  def __init__(self, projects : list[str], numRums : int, shouldClean : bool, enableAsan : bool, enableUbsan : bool, enableCodeCoverage : bool):
    self.projects = projects
    self.enable_asan = enableAsan
    self.enable_ubsan = enableUbsan
    self.enable_code_coverage = enableCodeCoverage
    self.should_clean = shouldClean
    self.num_runs = numRums
  
  def execute(self, singleThreaded : bool):
    regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
    
    # Generate the fuzzy tests
    rc = self._generate()
    _pass_results["fuzzy testing generation"] = rc
    if rc != 0:
      regis.diagnostics.log_err(f"failed to generate fuzzy code")
      return rc

    # pull out the generated projects, so we know what we can build and run
    test_projects_path = os.path.join(root_path, settings['intermediate_folder'], settings['build_folder'], 'test_projects.json')
    if not os.path.exists(test_projects_path):
      regis.diagnostics.log_err(f'"{test_projects_path}" does not exist.')
      return rc | 1

    # if no projects are specified, we run on all of them
    test_projects = regis.rex_json.load_file(test_projects_path)
    fuzzy_test_projects = CaseInsensitiveDict(test_projects["TypeSettings"].get("Fuzzy"))

    self.projects = self.projects or list(fuzzy_test_projects.keys())

    # If we still have no projects, generation silently failed.
    # either way, we need to exit here
    if not self.projects:
      regis.diagnostics.log_warn(f'No fuzzy test projects found. have you generated them?')
      _pass_results["fuzzy testing - nothing to do"] = rc
      return rc

    # Now build the projects we're interested in
    regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
    rc |= self._build(self.projects, singleThreaded)

    # if any builds fail we can't run any tests
    # so we exit here
    _pass_results["fuzzy testing building"] = rc
    if rc != 0:
      regis.diagnostics.log_err(f"failed to build fuzzy code")
      return rc

    # Now that we've build everything, let's run everything
    for project in self.projects:
      if project not in fuzzy_test_projects:
        regis.diagnostics.log_err(f'project "{project}" not found in {test_projects_path}. Please check its generation settings')
        continue

      # get the project test settings out of our testing files
      project_settings = fuzzy_test_projects[project]
      runnables = project_settings['TargetRunnables']
      working_dir = project_settings['WorkingDir']

      # run all the tests
      new_rc = self._run(runnables, working_dir)
      _pass_results[f'fuzzy tests result - {project}'] = rc

      rc |= new_rc

    # report any issues
    if rc != 0:
      regis.diagnostics.log_err('fuzzy tests failed')
      return rc
    
    return rc

  def _generate(self):
    with regis.task_raii_printing.TaskRaiiPrint("generating fuzzy testing code"):

      config_args = []
      config_args.append(f'-intermediate-dir={fuzzy_intermediate_dir}')
      config_args.append('-enable-fuzzy-testing')
      config_args.append('-disable-default-configs')

      if self.enable_code_coverage:
        config_args.append('-enable-code-coverage')

      if self.enable_asan:
        config_args.append('-enable-address-sanitizer')

      if self.enable_ubsan:
        config_args.append('-enable-ub-sanitizer')

      config = regis.generation.create_config(' '.join(config_args))
      return _generate_test_files(self.should_clean, fuzzy_intermediate_dir, config)
  
  def _build(self, projects : list[str], singleThreaded : bool):
    with regis.task_raii_printing.TaskRaiiPrint("building unit tests"):
      return _build_files(projects, singleThreaded)
  
  def _run(self, runnables : list, workingDir : str):
     with regis.task_raii_printing.TaskRaiiPrint("running unit tests"):
      with regis.util.temp_cwd(workingDir):

        rc = 0
    
        # loop over each unit test program path and run it
        for runnable_dict in runnables:
          args = []
          args.append('corpus')
          args.append(f'-runs={self.num_runs}')
          runnable = Runnable(runnable_dict, args, self.enable_asan, self.enable_ubsan)
          new_rc = runnable.run()

          if new_rc != 0:
            regis.diagnostics.log_err(f"fuzzy testing failed for {runnable.program}") # use full path to avoid ambiguity

          rc |= new_rc

        return rc

# the compdbPath directory contains all the files needed to configure clang tools
# this includes the compiler database, clang tidy config files, clang format config files
# and a custom generated project file, which should have the same filename as the source root directory
# of the project you're testing
def _get_project_name_of_compdb(compdbPath):
  dirs = os.listdir(compdbPath)
  for dir in dirs:
    if ".project" in dir:
      return dir.split(".")[0]
  
  return ""

def _find_files(folder, predicate):
  found_files : list[str] = []

  for root, dirs, files in os.walk(folder):
    for file in files:
      if predicate(file):
        path = os.path.join(root, file)
        found_files.append(path)      
  
  return found_files

def _generate_test_files(shouldClean : bool, intermediateDir : str, config):
  """Perform a generation for a test"""
  if shouldClean:
    # Clean the intermediates first if specified by the user
    # we clean in the generation step, to make sure that we only generate the unit tests we need
    full_intermediate_dir = _create_full_intermediate_dir(intermediateDir)
    regis.diagnostics.log_info(f"cleaning {full_intermediate_dir}..")
    regis.util.remove_folders_recursive(full_intermediate_dir)

  return regis.generation.new_generation(settings, config)

def _build_files(projectsToBuild : list[str] = "", singleThreaded : bool = False):
  """Build certain projects under a intermediate directory in certain configs using certain compilers
  This is useful after a generation to make sure all projects are build
  """

  res = 0

  # loop over the projects create a build for each combination
  for project in projectsToBuild:
    res |= regis.build.build_all_configs(project, shouldBuild=True, shouldClean=False, buildDependencies=True, singleThreaded=singleThreaded)

  return res

def _create_full_intermediate_dir(dir):
  """Create the absolute path for the test build directory"""
  return os.path.join(os.getcwd(), settings["intermediate_folder"], settings["build_folder"], dir)

def _find_tests_file(projectSettings : dict):
  project_root = projectSettings["Root"]
  test_file_path = os.path.join(project_root, "tests.json")

  if not os.path.exists(test_file_path):
    return None

  return test_file_path

# public API
def test_include_what_you_use(shouldClean : bool = True, singleThreaded : bool = False, shouldFix : bool = False):
  iwyu_job = IncludeWhatYouUseJob(shouldClean, shouldFix)
  return iwyu_job.execute(singleThreaded)
  
def test_clang_tidy(filesRegex = ".*", shouldClean : bool = True, singleThreaded : bool = False, filterLines : bool = False, autoFix : bool = False):
  clang_tidy_job = ClangTidyJob(shouldClean, autoFix, filterLines, filesRegex)
  return clang_tidy_job.execute(singleThreaded)

def test_unit_tests(projects, shouldClean : bool = True, singleThreaded : bool = False, enableAsan : bool = False, enableUbsan : bool = False, enableCoverage : bool = False):
  unit_test_job = UnitTestJob(projects, shouldClean, enableAsan, enableUbsan, enableCoverage)
  return unit_test_job.execute(singleThreaded)

def test_fuzzy_testing(projects, numRuns, shouldClean : bool = True, singleThreaded : bool = False, enableAsan : bool = False, enableUbsan : bool = False, enableCodeCoverage : bool = False):
  fuzzy_test_job = FuzzyTestJob(projects, numRuns, shouldClean, enableAsan, enableUbsan, enableCodeCoverage)
  return fuzzy_test_job.execute(singleThreaded)
  
def run_auto_tests(projects, timeoutInSeconds : int, shouldClean : bool = True, singleThreaded : bool = False, enableAsan : bool = False, enableUbsan : bool = False, enableCodeCoverage : bool = False):
  auto_test_job = AutoTestJob(projects, timeoutInSeconds, shouldClean, enableAsan, enableUbsan, enableCodeCoverage)
  return auto_test_job.execute(singleThreaded)

# Creating new projects
class TestProjectType(Enum):
  UnitTest = 0,
  AutoTest = auto(),
  FuzzyTest = auto(),

def create_new_project(solutionFolder : str, project : str, projectType : TestProjectType):
  if projectType == TestProjectType.UnitTest:
    sharpmake_file_template = os.path.join(root_path, '_build', 'sharpmake', 'templates', 'rex_unit_test_template.sharpmake.cs')
    file_template = os.path.join(root_path, '_build', 'sharpmake', 'templates', 'rex_unit_test_template.cpp')
  if projectType == TestProjectType.FuzzyTest:
    sharpmake_file_template = os.path.join(root_path, '_build', 'sharpmake', 'templates', 'rex_fuzzy_test_template.sharpmake.cs')
    file_template = os.path.join(root_path, '_build', 'sharpmake', 'templates', 'rex_fuzzy_test_template.cpp')
  if projectType == TestProjectType.AutoTest:
    sharpmake_file_template = os.path.join(root_path, '_build', 'sharpmake', 'templates', 'rex_auto_test_template.sharpmake.cs')
    file_template = os.path.join(root_path, '_build', 'sharpmake', 'templates', 'rex_auto_test_entry_template.cpp')

  # first make the directory that'll hold all the project's source files
  project_folder = regis.util.to_snakecase(project)
  project = regis.util.to_camelcase(project)
  project_dir = os.path.join(root_path, 'tests', solutionFolder, project_folder)
  if os.path.isdir(project_dir):
    regis.diagnostics.log_err(f'project directory "{project_dir}" already exists.')
    return

  os.makedirs(project_dir)
  regis.diagnostics.log_info(f'project dir created at {project_dir}')

  # create the sub directories for the source files
  os.mkdir(os.path.join(project_dir, 'include'))
  os.mkdir(os.path.join(project_dir, 'config'))
  os.mkdir(os.path.join(project_dir, 'src'))

  # create the sharpmake file
  test_sharpmake_file = os.path.join(project_dir, f'{project_folder}.sharpmake.cs')
  sharpmake_content = open(sharpmake_file_template).read()
  sharpmake_content = sharpmake_content.replace('<UnitTestProjectName>', project)
  sharpmake_content = sharpmake_content.replace('<test_solution_folder>', solutionFolder)
  open(test_sharpmake_file, 'w').write(sharpmake_content)

  # copy the test template
  test_file = os.path.join(project_dir, 'src', f'{project_folder}.cpp')
  shutil.copy(file_template, test_file)  

  # for auto tests we need to copy a template of an auto test file as well
  if projectType == TestProjectType.AutoTest:
    file_template = os.path.join(root_path, '_build', 'sharpmake', 'templates', 'rex_auto_test_template.cpp')
    auto_test_file = os.path.join(project_dir, 'src', f'{project_folder}_test.cpp')
    shutil.copy(file_template, auto_test_file)  



import os
import json

class CompileDBEntry:
  def __init__(self, jsonEntry):
    self.directory = jsonEntry['directory']
    self.command = jsonEntry['command']
    self.file = jsonEntry['file']
    self.output = jsonEntry['output']

class CompileDB:
  def __init__(self, path):
    database = json.load(open('compile_commands.json'))
    self.entries = list([CompileDBEntry(entry)
           for entry in database])

db = CompileDB('compile_commands.json')
for entry in db.entries:
  print(entry.file)